program OthKonv;

{$APPTYPE CONSOLE}

uses
  SysUtils,Classes;

const richtung:array[1..8] of integer = (1,10,9,11,-1,-10,-9,-11);
Const PassMove     = 128;

type TMoveArray = Array[1..60] of byte;

type pulia = (Leer,Schwarz,Weiss,Edge); // Steine auf dem Board
type TBoard = Array[0..99] of Pulia;
type TBoard64 = packed Array[1..64] of Pulia;

type TShortMove = Record
                   Move:byte;
                   Score:Integer;
                  end;

type TGameInfo = RECORD
                  BlackName,WhiteName:String;      
                  Result:ShortInt;
                  Anzahl:Byte;
                  Start_Board : TBoard64;
                  Moves:Array[1..60] of TShortMove;
                 END;

type WTHORHeader = Record
                    Century     :Byte;
                    Year        :Byte;
                    Month       :Byte;
                    Day         :Byte;
                    Anzahl      :Cardinal;
                    TourAnz     :Word;
                    YearOfGames :Word;
                    SizeOfBoard :Byte;
                    IsAnEndgameSolitaireFile : Byte;
                    NumberOfEmptiesForTheoricalScoreCalculation  : Byte;
                    Reserved    :Byte;
                   end;

type WTHORRecord = Record
                    TourNumber : Word;
                    Black      : Word;
                    White      : Word;
                    RealScore  : Byte;
                    TheoryScore: Byte;
                    MoveList   : TMoveArray;
                   end;

var WThorFile:File of Byte;
var Count,LocalCount:Integer;
var PlayerList:TStringList;
var SearchRec:TSearchRec;
var FileName:String;

Procedure WriteWthorHeader(Count:Integer);
var Header:WThorHeader;
begin
 Header.Century := 20;
 Header.Year := 02;
 Header.Month := 1;
 Header.Day := 1;
 Header.Anzahl := Count;
 Header.TourAnz := 0;
 Header.YearOfGames := 2002;
 Header.SizeOfBoard := 8;
 Header.IsAnEndgameSolitaireFile := 0;
 Header.NumberOfEmptiesForTheoricalScoreCalculation := 1;
 Header.Reserved := 0;
 BlockWrite(WThorFile,Header,sizeof(Header));
end;

Procedure KonvertGame(const aGame:TGameInfo);
var Rec:WThorRecord;
var i,BlackStones:Integer;
begin
 Rec.TourNumber := 0;
 Rec.Black := PlayerList.IndexOf(aGame.BlackName);
 Rec.White := PlayerList.IndexOf(aGame.WhiteName);
 Rec.RealScore := aGame.Result;
 Rec.TheoryScore := aGame.Result;
 for i := 1 to aGame.Anzahl do
  Rec.MoveList[i] := aGame.Moves[i].Move;
 BlockWrite(WThorFile,Rec,sizeof(Rec));
end;

Function RotateMove(Move,Art:Byte):Byte;
var x,y,Temp:Byte;
begin
result := 0;
case Art of
     0:result := Move;
     1: begin
         // 56 to 34
         x := Move div 10;
         y := Move mod 10;
         Temp := x;
         x := y;
         y := Temp;
         result := (9-x)*10+(9-y);
        end;
     2: begin
         // 43 to 34
         x := Move div 10;
         y := Move mod 10;
         result := y*10+x;
        end;
     3:begin
         // 65 to 34
         Move := RotateMove(Move,2);
         Move := RotateMove(Move,1);
         result := Move;
       end;
end;
end;

Procedure KonvertMovesTo54(var GameRec:TGameInfo);
var rotmove,i,rot:Integer;
var RotTo56:Integer;
begin
 RotTo56 := -1;
 for rot := 0 to 3 do
 begin
  rotmove := RotateMove(GameRec.Moves[1].Move,Rot);
  if rotmove=56 then
  begin
   RotTo56 := rot;
   break;
  end;
 end;
 if RotTo56 = -1 then
 begin
  Writeln('ERROR: I could not konvert moves to E5!');
  RotTo56 := 0;
 end;
 for i := 1 to 60 do
 begin
  GameRec.Moves[i].Move := RotateMove(GameRec.Moves[i].Move,RotTo56);
 end;

end;

Function OtherSite(Pl:Pulia):Pulia;
begin
 if Pl = Weiss then result := Schwarz else result := Weiss;
end;

Procedure PlayMove(var Board:TBoard;Player,Gegner:Pulia;Wohin:Byte);
var Geg:Pulia;
var ri,Pos:Integer;
var HatGespielt:Boolean;
begin
 Geg := Gegner;
 HatGespielt := False;
 for ri := 1 to 8 do
 begin
   pos := Wohin+Richtung[ri];
   while (Board[Pos] = Geg) do
    Inc(pos,Richtung[ri]);
   if Board[Pos] = Player then
   begin
    pos := Wohin+Richtung[ri];
    while Board[Pos] = Gegner do
    begin
     HatGespielt := true;
     Board[Pos] := Player;
     Inc(pos,Richtung[ri]);
    end; {while b[Pos]}
  end;
 end;
 if Hatgespielt then
  Board[Wohin] := Player;
end;

Function Mobility(const Board:TBoard;Pl:Pulia):Integer;
var i,ri,Pos:Integer;
var Geg:Pulia;
begin
 Geg := OtherSite(Pl);
 result := 0;
 for i := 11 to 88 do
 begin
  if Board[i] = Leer then
  begin
   for ri := 1 to 8 do
   begin
    Pos := i+Richtung[ri];
    if (Board[Pos] <> Geg) then Continue;
    Inc(Pos,Richtung[ri]);
    while (Board[Pos] = Geg) do
     Inc(Pos,Richtung[ri]);
    if (Board[Pos] = Pl) then
    begin
     INC(result);
     break;
    end;
   end;
  end;
 end;
end;

Function BlackStones(const Board:TBoard):Integer;
var sch,wei,i:Integer;
begin
 sch := 0;
 wei := 0;
 for i := 11 to 88 do
 begin
  if Board[i] = Schwarz then inc(sch);
 end;
 result := sch;
end;

Procedure ReadBefehl(aGame,Bef:String;var Value:String);
var p1:Integer;
var i:Integer;
begin
 Value := '';
 P1 := System.Pos(Bef+'[',aGame);
 if P1 = 0 then exit;
 i := P1+Length(Bef)+1;
 repeat
  if aGame[i]<>']' then Value := Value+aGame[i];
  INC(i);
 until aGame[i] = ']';
end;

Function Is8x8Standard(aGame:String):Boolean;
var Value,myStr:String;
var anzRand,myPos:Integer;
begin
 ReadBefehl(aGame,'TY',Value);
 result := Value = '8';
end;

Function IsSynchroRand(aGame:String):Boolean;
var Value,myStr:String;
var anzRand,myPos:Integer;
begin
 ReadBefehl(aGame,'TY',Value);
 myPos := System.Pos('s8r',Value);
 anzRand := 0;
 if myPos>0 then
 begin
  myStr := Copy(Value,myPos+3,2);
  anzRand := StrToInt(myStr);
 end;
 result := (anzRand>0) and (anzRand mod 2 = 0);
end;

function IsTopPlayer(aPlayer:String):Boolean;
begin
 aPlayer := UpperCase(aPlayer);
 if (aPlayer = 'PIGLET') or
    (aPlayer = 'KITTY') or
    (aPlayer = 'LYNX') or
    (aPlayer = 'CAESAR') or
    (aPlayer = 'CHARLY+') or
    (aPlayer = 'NTEST') or
    (aPlayer = 'EDAX') or
    (aPlayer = 'VORAX') or
    (aPlayer = 'EDAX12') or
    (aPlayer = 'YOHOHO') or
    (aPlayer = 'YOHO') or
    (aPlayer = 'HAWK12') or
    (aPlayer = 'CONDOR') or
    (aPlayer = 'BOT_TEST') or
    (aPlayer = 'HAWK') or
    (aPlayer = 'BREVERSI') or
    (aPlayer = 'CYBOT') or
    (aPlayer = 'MOUSE') or
    (aPlayer = 'BOPEN') or
    (aPlayer = 'BOSCO') or
    (aPlayer = 'BTEST') or
    (aPlayer = 'SOFIA') or
    (aPlayer = 'CHARLY') or
    (aPlayer = 'S16') or
    (aPlayer = 'OO7') or
    (aPlayer = 'ZORRO') or
    (aPlayer = 'SLARIN') or
    (aPlayer = 'BOOKLET') or
    (aPlayer = 'BOOKLET+') or
    (aPlayer = 'VO') or
    (aPlayer = 'REINDEER') or
    (aPlayer = 'ZEBRA') or
    (aPlayer = 'VO+') then result := true else result := False;
end;

Procedure addGGFGame(aGame:String;Var GameFinished:Boolean;OnlyCount:Boolean);
var aBoard:TBoard;
var ResultStr,strBoard,BlackPlayer,WhitePlayer:String;
var toPlay:Pulia;
var GameRec:TGameInfo;
var Code,GameResult:Integer;
var theMove : TShortMove;

Function ParseMove(MoveStr:String):TShortMove;
var myMove : String;
var m:Byte;
var p1:Integer;
var ch:Char;
begin
 p1 := System.Pos('/',MoveStr);
 if p1 = 0 then myMove := UpperCase(MoveStr) else
  myMove := UpperCase(Copy(MoveStr,1,p1-1));
 if System.Pos('PA',myMove)>0 then m := PassMove else
 begin
  ch := myMove[1];
  m := (ord(ch)-ord('A')+1)+StrToInt(myMove[2])*10;
 end;
 result.Move := m;
 Result.Score := 0;
end;

Procedure ParseMoves(aGame:String;var GameFinished:Boolean);
var MoveStr:String;
var nextPlayer,Player,Gegner,temp:Pulia;
var p1:Integer;
var testP1,PosInMove:Integer;
var Board:TBoard;
var LastPassMove:Boolean;
begin
 GameFinished := False;
// OthelloEngine.InitStartPosition(aBoard,toPlay,OthelloEngine.otherSite(toPlay));
 Board := aBoard;
 nextPlayer := toPlay;
 Player := toPlay;
 Gegner := OtherSite(Player);
 PosInMove := 1;
 LastPassMove := false;
 repeat
  if nextPlayer = Schwarz then
  begin
   ReadBefehl(aGame,']B',MoveStr);
   if MoveStr = '' then break;
   nextPlayer := Weiss;
   p1 := System.Pos(']B[',aGame);
   aGame := Copy(aGame,p1+3+Length(MoveStr),Length(aGame)-p1-2-Length(MoveStr));
  end else
  begin
   ReadBefehl(aGame,']W',MoveStr);
   if MoveStr = '' then break;
   nextPlayer := Schwarz;
   p1 := System.Pos(']W[',aGame);
   aGame := Copy(aGame,p1+3+Length(MoveStr),Length(aGame)-p1-2-Length(MoveStr));
  end;
  theMove := ParseMove(MoveStr);
  if theMove.Move<>PassMove then
  begin
   PlayMove(Board,Player,Gegner,theMove.Move);
   Temp := Player;
   Player := Gegner;
   Gegner := Temp;
   GameRec.Moves[PosInMove] := theMove;
   GameRec.Anzahl := PosInMove;
   INC(PosInMove);
   LastPassMove := false;
  end else
  Begin
   Temp := Player;
   Player := Gegner;
   Gegner := Temp;
   LastPassMove := True;
  end;
 until aGame = '';
 if Mobility(Board,Player)=0 then
 begin
   GameFinished := Mobility(Board,Gegner)=0;
 end;
 if GameFinished then
  GameRec.Result := BlackStones(Board);
end;

Function ParseBoard(aBoard:String;var toPlay:pulia):TBoard;
var BoardType,S1:String;
var p1,i,Pos88:Integer;
var TypeVal,Code:Integer;
begin
 p1 := System.Pos(' ',aBoard);
 BoardType := Copy(aBoard,1,p1-1);
 val(BoardType,TypeVal,Code);
 if TypeVal<>8 then exit;
 S1 := Copy(aBoard,Length(aBoard),1);
 if S1 = '*' then toPlay := Schwarz else toPlay := Weiss;
 aBoard := Copy(aBoard,p1+1,Length(aBoard)-p1-2);
 for i := 0 to 99 do Result[i] := Edge;
 Pos88 := 11;
 for i := 1 to 72 do
 begin
  S1 := copy(aBoard,i,1);
  if S1 = '-' then Result[Pos88] := Leer;
  if S1 = '*' then Result[Pos88] := Schwarz;
  if S1 = 'O' then Result[Pos88] := Weiss;
  if S1 = ' ' then INC(Pos88);
  INC(Pos88);
 end;
end;

begin
 ReadBefehl(aGame,'PB',BlackPlayer);
 ReadBefehl(aGame,'PW',WhitePlayer);
 GameRec.BlackName := BlackPlayer;
 GameRec.WhiteName := WhitePlayer;

 ReadBefehl(aGame,'BO',strBoard);
 aBoard := ParseBoard(strBoard,toPlay);
 ReadBefehl(aGame,'RE',ResultStr);
 val(ResultStr,GameResult,Code);
 GameRec.Result := GameResult;
 ParseMoves(aGame,GameFinished);
 if not OnlyCount and GameFinished then
 begin
  KonvertMovesTo54(GameRec);
  if GameRec.Moves[1].Move<>56 then
   Code := Round(sin(1));
  KonvertGame(GameRec);
 end else
 begin
  if PlayerList.IndexOf(BlackPlayer)=-1 then
   PlayerList.add(BlackPlayer);
  if PlayerList.IndexOf(WhitePlayer)=-1 then
   PlayerList.add(WhitePlayer);
 end;
end;

Function CheckGame(aGame:String):Boolean;
var Value:String;
var BlackPlayer,WhitePlayer:String;
var addGameBool:Boolean;
begin
 result := False;
 ReadBefehl(aGame,'GM',Value);
 if UpperCase(Value) <> 'OTHELLO' then exit;
 addGameBool := {IsSynchroRand(aGame) or} is8x8Standard(aGame);
 if not addGameBool then exit;
 ReadBefehl(aGame,'PB',BlackPlayer);
 ReadBefehl(aGame,'PW',WhitePlayer);
// if IsTopPlayer(BlackPlayer) and IsTopPlayer(WhitePlayer) then
  result := true;
end;

Function ReadGGFFile(FileName:String;OnlyCount:Boolean):Integer;
var Count:Integer;
var InFile:Text;
var GameFinished:Boolean;
var aGame,aLine:String;

Function ReadLine:String;
var s:String;
var ch:Char;
const EOL = chr(10);
begin
 s := '';
 repeat
  read(InFile,Ch);
  if Ch<>EOL then s := s+Ch;
 until Eof(InFile) or (ch = EOL);
 result := s;
end;

Function ReadGame(var aGame:String):String;
var p1,p2:Integer;
begin
 p1 := System.Pos('(;',aGame);
 p2 := System.Pos(';)',aGame);
 Result := Copy(aGame,p1+2,p2-p1-2);
 aGame := Copy(aGame,p2+2,Length(aGame)-p2);
end;

begin
 AssignFile(InFile,FileName);
 reset(InFile);
 Count := 0;
 While not Eof(inFile) do
 begin
  aLine := ReadLine;
  repeat
   aGame := ReadGame(aLine);
   if CheckGame(aGame) then
   begin
    addGGFGame(aGame,GameFinished,OnlyCount);
    if GameFinished then INC(Count);
   end;
  until aLine='';
 end;
 CloseFile(InFile);
 result := Count;
end;

Procedure LoadWThorBase(FileName:String);
var Header:WThorHeader;
var Rec:WThorRecord;
var InFile:File of byte;
var GameNr:Integer;
begin
 AssignFile(InFile,FileName);
 reset(InFile);
 BlockRead(InFile,Header,sizeof(WTHORHeader));   // Header einlesen
 for GameNr := 1 to Header.Anzahl do
 begin
  BlockRead(InFile,Rec,sizeof(WTHORRecord));   // Game einlesen
 end;
 CloseFile(InFile);
end;

Procedure LoadWThorTrn(FileName:String);
var Header:WThorHeader;
var Rec:WThorRecord;
var InFile:File of byte;
var GameNr:Integer;
var TournamentName:Array[1..26] of Char;
begin
 AssignFile(InFile,FileName);
 reset(InFile);
 BlockRead(InFile,Header,sizeof(WTHORHeader));   // Header einlesen
 for GameNr := 1 to Header.TourAnz do
 begin
  BlockRead(InFile,TournamentName,Sizeof(TournamentName));
 end;
 CloseFile(InFile);
end;

Procedure LoadWThorJou(FileName:String);
var Header:WThorHeader;
var Rec:WThorRecord;
var InFile:File of byte;
var GameNr:Integer;
//var Header:Array[0..15] of Byte;
var PlayerName:Array[1..20] of Char;
begin
 AssignFile(InFile,FileName);
 reset(InFile);
 BlockRead(InFile,Header,sizeof(WTHORHeader));   // Header einlesen
 for GameNr := 1 to Header.TourAnz do
 begin
  BlockRead(InFile,PlayerName,Sizeof(PlayerName));
 end;
 CloseFile(InFile);
end;

Procedure WriteWThorJou(FileName:String);
var WThorJou:File of Byte;
var Header:WThorHeader;
var i,k:Integer;
var Player:String;
var CharPlayer:Array[1..20] of Char;
begin
 Assign(WThorJou,FileName);
 Rewrite(WThorJou);
 
 Header.Century := 20;
 Header.Year := 02;
 Header.Month := 12;
 Header.Day := 20;
 Header.Anzahl := 0;
 Header.TourAnz := PlayerList.Count;
 Header.YearOfGames := 0;
 Header.Reserved := 0;
 
 BlockWrite(WThorJou,Header,sizeof(Header));

 for i := 0 to PlayerList.Count-1 do
 begin
  FillChar(CharPlayer,Sizeof(CharPlayer),chr(0));
  Player := PlayerList[i];
  for k := 1 to Length(Player) do
   CharPlayer[k] := Player[k];
  BlockWrite(WThorJou,CharPlayer,sizeof(CharPlayer)); 
 end;
 
 CloseFile(WThorJou);
end;

begin
  { TODO -oUser -cConsole Main : Hier Code einfgen }

// LoadWthorJou('WTHOR.Jou');
// LoadWthorTrn('WTHOR.TRN');
// LoadWthorBase('WTH_1980.wtb');
 PlayerList := TStringList.Create;
 PlayerList.Add('???');
 Assign(WThorFile,'GGS_2002.wtb');
 Rewrite(WThorFile);

 Count := 0;
 Writeln('Analysing GGS databases (8x8 standard)');
 if FindFirst('*.ggf',$3F,SearchRec)=0 then
 begin
  repeat
   FileName := SearchRec.Name;
   LocalCount := ReadGGFFile(FileName,True);
   INC(Count,LocalCount);
   Writeln('     GGS Database (',FileName,'): ',localCount,' games');
  until FindNext(SearchRec)<>0;
  FindClose(SearchRec);
 end;
 Writeln('Exporting #',Count,' GGS games (8x8 standard) to WTHOR database');

 WriteWThorHeader(Count);

 Count := 0;
 if FindFirst('*.ggf',$3F,SearchRec)=0 then
 begin
  repeat
   FileName := SearchRec.Name;
   LocalCount := ReadGGFFile(FileName,False);
   INC(Count,LocalCount);
   Writeln('     Exported GGS Database (',FileName,'): ',localCount,' games');
  until FindNext(SearchRec)<>0;
  FindClose(SearchRec);
 end;
 Writeln(Count,' of games have been exported');

 WriteWThorJou('GGS.JOU');

 PlayerList.Free;
end.
